#ifndef FIELD_HPP_INCLUDED
#define FIELD_HPP_INCLUDED

#include "main.hpp"

/** 
 * Field types
 */
/// Cell free for turret building
const int FIELD_FREE = 0;
/// Turret on the cell
const int FIELD_TURRET = 1;
/// Cell is part of the mob path
const int FIELD_PATH = 2;
/// Cell is blocked
const int FIELD_BLOCKED = 3;

struct FieldCellCoord{
    int x, y;
    
    FieldCellCoord( int x = 0 , int y = 0 );
};

struct FieldCell{
    int type;
    int64 turret;
    
    FieldCell( int type = 0 , int64 turret = 0LL );
    
    void setVisibleBy( int turret_id );
    void setInvisibleBy( int turret_id );
    bool isVisibleBy( int turret_id ) const;
};

struct Field{
    /**
     * The size of the field
     */
    int rows;
    int cols;
    /**
     * The field itself
     */
        /**
         * The types of each field cell
         */
        /**
         * Description of the turrets
         * For FIELD_PATH shows which battle
         *  turrets see this cell
         * For FIELD_TURRET shows which
         *  turret is built there
         * For FIELD_BLOCKED and FIELD_FREE
         *  shows nothing
         */
    std::vector< std::vector< FieldCell > > cell;

        /**
         * Description of the mob path
         */
    std::vector< FieldCellCoord > path;
    
    Field( int rows , int cols , const char field[MAX_FIELD_SIZE][MAX_FIELD_SIZE] );
    
    bool isInside( int x , int y ) const;
    bool placeTurret( int x , int y , int id );
    bool removeTurret( int x , int y );
    bool isFree( int x , int y ) const;
    bool isTurret( int x , int y ) const;
    bool isPath( int x , int y ) const;
    bool isBlocked( int x , int y ) const;
    int getTurretId( int x , int y ) const;
    int64 getTurretsSeeing( int x , int y ) const;
    
    /**
     * Some wrapper methods around FieldCell methods
     */
    
    bool setVisibleBy( int x , int y , int turret_id );
    bool setInvisibleBy( int x , int y , int turret_id );
    bool isVisibleBy( int x , int y , int turret_id ) const;
};

#endif // FIELD_HPP_INCLUDED
