#ifndef GAME_HPP_INCLUDED
#define GAME_HPP_INCLUDED

#include "main.hpp"
#include "Field.hpp"
#include "Mob.hpp"
#include "PlacedTurret.hpp"
#include "RunningMob.hpp"
#include "Quadtree.hpp"
#include "Segmenttree.hpp"
#include "WaveRunner.hpp"

void defaultMobInRangeForOneCall( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument );
void defaultMobInRangeForManyCalls( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument );
void defaultMobFellDead( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument );
void defaultMobReachedTheEnd( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument );

struct Game{
    /** 
     * The field
     */
    Field *field;
    /**
     * The turrets ( turrets.size() == MAX_TURRET_COUNT )
     */
    std::vector< PlacedTurret* > turrets;
    /**
     * The quadtrees for each modifier tower type
     */
    Quadtree *damage_mod[MAX_MOB_TYPES];
    Quadtree *rate_mod[MAX_MOB_TYPES];
    /**
     * Fast visibility info on top Field's O(1) array based
     */
    Segmenttree *visibilityTree;
    
    /**
     * The player info
     */
    
    /**
     * The lives left
     */
    int lives;
    /**
     * Mobs killed
     */
    int kills;
    /**
     * Money
     */
    int64 money;
    /**
     * Next wave number
     */
    int wave;
    
    Game( int rows , int cols , const char field[MAX_FIELD_SIZE][MAX_FIELD_SIZE] );
    
    PlacedTurret *getTurret( int turret_id ) const;
    
    int buyTurret( const Turret *turret , int x , int y );
    bool sellTurret( int x , int y , bool remove_only = false );
    
    void go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs );
    
    void go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs ,
             WaveRunner_Callback_Type myMobInRange ,
             WaveRunner_Callback_Type myMobFellDead ,
             WaveRunner_Callback_Type myMobReachedTheEnd ,
             void *argument
        );
    
    void go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs ,
             WaveCompleted_Callback_Type waveCompleted ,
             WaveRunner_Callback_Type myMobInRange ,
             WaveRunner_Callback_Type myMobFellDead ,
             WaveRunner_Callback_Type myMobReachedTheEnd ,
             void *argument
        );
        
    int totalTurretsBuild() const;
    
		// sod
    int _turrets_built;
    protected:
    
    int64 _turret_mask;
};

#endif // GAME_HPP_INCLUDED
