#include "Quadtree.hpp"

static bool isBetween( int l , int r , int x ){
    return l <= x and x <= r;
}

static bool rectInsideRect( int upxi , int upyi , int lowxi , int lowyi ,
                            int upxj , int upyj , int lowxj , int lowyj )
{
    return isBetween( upxi , lowxi , upxj ) and isBetween( upxi , lowxi , lowxj )
        and isBetween( upyi , lowyi , upyj ) and isBetween( upyi , lowyi , lowyj );
}

Quadtree::Quadtree( int size , double init )
    :
        size(size), data(), self()
{
    data.resize(1 + 2 * size * size);
    self.resize(1 + 2 * size * size);
    
    _build( 1 , 0 , 0 , size , size , init );
}

void Quadtree::_build( int at , int upper_left_x , int upper_left_y , int lower_right_x , int lower_right_y , double value ){
    int height = (lower_right_x - upper_left_x);
    int width = (lower_right_y - upper_left_y);
    
    self[at] = 0.0;
    
    if( upper_left_x == lower_right_x - 1 and upper_left_y == lower_right_y - 1 ){
        data[at] = self[at] = value;
    
        return;
    }
    
    int midx = (upper_left_x + lower_right_x) / 2;
    int midy = (upper_left_y + lower_right_y) / 2;
    
    _build( at * 4 , upper_left_x , upper_left_y , midx , midy , value );
    _build( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , value );
    _build( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , value );
    _build( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , value );
    
    data[at] = self[at] * (width * height)
        + data[at * 4] + data[at * 4 + 1]
        + data[at * 4 + 2] + data[at * 4 + 3];
}

double Quadtree::query( int upper_left_x , int upper_left_y , int lower_right_x , int lower_right_y ) const{
    return _query( 1 , 0 , 0 , size , size , upper_left_x , upper_left_y , lower_right_x , lower_right_y );
}

void Quadtree::update( int upper_left_x , int upper_left_y , int lower_right_x , int lower_right_y , double value ){
    _update( 1 , 0 , 0 , size , size , upper_left_x , upper_left_y , lower_right_x , lower_right_y , value );
}

void Quadtree::_update( int at , 
    int upper_left_x , int upper_left_y , int lower_right_x , int lower_right_y ,
    int my_upper_left_x , int my_upper_left_y , int my_lower_right_x , int my_lower_right_y ,
    double value )
{
    int width = lower_right_y - upper_left_y;
    int height = lower_right_x - upper_left_x;
 
    if( upper_left_x == my_upper_left_x and upper_left_y == my_upper_left_y
       and lower_right_x == my_lower_right_x and lower_right_y == my_lower_right_y ){
        self[at] += value;
        
        data[at] += value * (double) (width * height);
        
        return;
    }
    
    int midx = (lower_right_x + upper_left_x) / 2;
    int midy = (lower_right_y + upper_left_y) / 2;
    
    /// Oh man!
    
    if( rectInsideRect( upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        _update( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    else if( rectInsideRect( upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        _update( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    else if( rectInsideRect( midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        _update( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    else if( rectInsideRect( midx , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        _update( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    else if( rectInsideRect( upper_left_x , upper_left_y , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        _update( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , midy , value );
        _update( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , midy , my_lower_right_x , my_lower_right_y , value );
    }else if( rectInsideRect( midx , upper_left_y , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        _update( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , midy , value );
        _update( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , my_upper_left_x , midy , my_lower_right_x , my_lower_right_y , value );
    }else if( rectInsideRect( upper_left_x , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        _update( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , midx , my_lower_right_y , value );
        _update( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , midx , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    }else if( rectInsideRect( upper_left_x , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        _update( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , midx , my_lower_right_y , value );
        _update( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , midx , my_upper_left_y , my_lower_right_x , my_lower_right_y , value );
    }else{
        /// Whew! if there are no bugs so far....
        _update( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , midx , midy , value );
        _update( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , midy , midx , my_lower_right_y , value );
        _update( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , midx , my_upper_left_y , my_lower_right_x , midy , value );
        _update( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , midx , midy , my_lower_right_x , my_lower_right_y , value );
    }
    
    data[at] = self[at] * (width * height)
        + data[at * 4] + data[at * 4 + 1]
        + data[at * 4 + 2] + data[at * 4 + 3];
}

double Quadtree::_query( int at , 
    int upper_left_x , int upper_left_y , int lower_right_x , int lower_right_y ,
    int my_upper_left_x , int my_upper_left_y , int my_lower_right_x , int my_lower_right_y
    ) const
    
{
    if( upper_left_x == my_upper_left_x and upper_left_y == my_upper_left_y 
        and lower_right_x == my_lower_right_x and lower_right_y == my_lower_right_y )
        return data[at];
    
    int width = (my_lower_right_y - my_upper_left_y);
    int height = (my_lower_right_x - my_upper_left_x);
    
    int midx = (lower_right_x + upper_left_x) / 2;
    int midy = (lower_right_y + upper_left_y) / 2;
    
    double result = self[at] * (width * height);
    
    if( rectInsideRect( upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        result += _query( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y);
    else if( rectInsideRect( upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        result += _query( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y );
    else if( rectInsideRect( midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        result += _query( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y );
    else if( rectInsideRect( midx , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) )
        result += _query( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y );
    else if( rectInsideRect( upper_left_x , upper_left_y , midx , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        result += _query( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , midy );
        result += _query( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , midy , my_lower_right_x , my_lower_right_y );
    }else if( rectInsideRect( midx , upper_left_y , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        result += _query( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , midy );
        result += _query( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , my_upper_left_x , midy , my_lower_right_x , my_lower_right_y );
    }else if( rectInsideRect( upper_left_x , upper_left_y , lower_right_x , midy , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        result += _query( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , midx , my_lower_right_y );
        result += _query( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , midx , my_upper_left_y , my_lower_right_x , my_lower_right_y );
    }else if( rectInsideRect( upper_left_x , midy , lower_right_x , lower_right_y , my_upper_left_x , my_upper_left_y , my_lower_right_x , my_lower_right_y ) ){
        result += _query( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , my_upper_left_y , midx , my_lower_right_y );
        result += _query( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , midx , my_upper_left_y , my_lower_right_x , my_lower_right_y );
    }else{
        /// Whew! if there are no bugs so far....
        result += _query( at * 4 , upper_left_x , upper_left_y , midx , midy , my_upper_left_x , my_upper_left_y , midx , midy );
        result += _query( at * 4 + 1 , upper_left_x , midy , midx , lower_right_y , my_upper_left_x , midy , midx , my_lower_right_y );
        result += _query( at * 4 + 2 , midx , upper_left_y , lower_right_x , midy , midx , my_upper_left_y , my_lower_right_x , midy );
        result += _query( at * 4 + 3 , midx , midy , lower_right_x , lower_right_y , midx , midy , my_lower_right_x , my_lower_right_y );
    }
    
    return result;
}

