#include "Segmenttree.hpp"

Segmenttree::Segmenttree( int size , aggregate_fn_type aggregate , aggregate_fn_type updatefn , int64 init )
    :
        size(size), aggregate(aggregate), updatefn(updatefn), data()
{
    data.resize( 1 + 2 * size );
    
    std::fill( data.begin() , data.end() , init );
}

void Segmenttree::update( int x , int64 value ){
    _update( 1 , 0 , size , x , value );
}

int64 Segmenttree::query( int l , int r ) const{
    return _query( 1 , 0 , size , l , r );
}

void Segmenttree::_update( int at , int l , int r , int x , int64 value ){
    if( l == r - 1 ){
        data[at] = updatefn( data[at] , value );
        
        return;
    }
    
    int mid = ( l + r ) / 2;
    
    if( x < mid )
        _update( at * 2  , l , mid , x , value );
    else if( mid <= x )
        _update( at * 2 + 1 , mid , r , x , value );
    
    data[at] = aggregate( data[at * 2] , data[at * 2 + 1] );
}

int64 Segmenttree::_query( int at , int l , int r , int a , int b ) const{
    if( l == a and r == b )
        return data[at];
    
    int mid = (l + r) / 2;
    
    if( b <= mid )
        return _query( at * 2 , l , mid , a , b );
    else if( mid <= a )
        return _query( at * 2 + 1 , mid , r , a , b );
    else
        return aggregate(
                _query( at * 2 , l , mid , a , mid ),
                _query( at * 2 + 1 , mid , r , mid , b )
            );
}
