#include "TextGameReader.hpp"

TextGame *TextGameReader::read( const char *input_file_name ){
    static char field_buffer[MAX_FIELD_SIZE][MAX_FIELD_SIZE];
    static char turret_name_buffer[MAX_TURRET_NAME_LENGTH + 1];
    
    FILE *in = fopen( input_file_name , "r" );
    
    if( !in )
        return NULL;
    
    int rows, cols;
    
    if( fscanf( in , "%d %d" , &rows , &cols ) != 2 )
        return NULL;
    
    for( int i = 0 ; i < rows ; i++ ){
        if( fscanf( in , "%s" , field_buffer[i] ) != 1 )
            return NULL;
        
        for( int j = 0 ; j < cols ; j++ )
            if( !strchr( ".#xSE" , field_buffer[i][j] ) )
                return NULL; /// invalid field
    }
    
    TextGame *result = new TextGame( rows , cols , field_buffer );
    
    int turret_count;
    
    if( fscanf( in , "%d" , &turret_count ) != 1 )
        return NULL;
    
    if( turret_count < 0 )
        return NULL;
    
    for( int i = 0 ; i < turret_count ; i++ ){
        /**
         * Read a turret
         */
        if( fscanf( in , "%s" , turret_name_buffer ) != 1 )
            return NULL;
        
        Turret *newTurret = new Turret;
        
        newTurret->name = turret_name_buffer;
        
        if( fscanf( in , "%d %d %d %d %d" , 
                &newTurret->type,
                &newTurret->rate,
                &newTurret->range,
                &newTurret->buy_price,
                &newTurret->sell_price ) != 5 )
                    return NULL;
        
        /**
         * ? add some tests on whether the type
         *  is one of the allowed ones, etc, etc?
         */
        
        for( int j = 0 ; j < MAX_MOB_TYPES ; j++ )
            if( fscanf( in , "%lf" , &newTurret->affect[j] ) != 1 )
                return NULL;
        
        /// Read the ids of which this turret is an upgrade of
        
        int prereqs;
        
        if( fscanf( in , "%d" , &prereqs ) != 1 )
            return NULL;
        
        result->upgrade_graph.push_back( std::vector< int >() );
        
        for( int j = 0 ; j < prereqs ; j++ ){
            int prereq_id;
            
            if( fscanf( in , "%d" , &prereq_id ) != 1 )
                return NULL;
            
            /// WARNING: Assuming the graph is a DAG
            /// Maybe implement a DFS to check?
            
            assert( prereq_id >= 0 and prereq_id < turret_count );
            
            result->upgrade_graph[i].push_back( prereq_id );
        }
        
        result->all_turrets.push_back( newTurret );
    }
    
    /// Read the mobs
    
    int mob_count;
    
    if( fscanf( in , "%d" , &mob_count ) != 1 )
        return NULL;
    
    if( mob_count < 0 )
        return NULL;
    
    for( int i = 0 ; i < mob_count ; i++ ){
        /**
         * Read the mobs one by one
         */
        Mob *newMob = new Mob;
        
        if( fscanf( in , "%d %d %lf" , &newMob->type, &newMob->speed, &newMob->health ) != 3 )
            return NULL;
        
        result->all_mobs.push_back( newMob );
    }
    
    /// Read lives and money
    if( fscanf( in , "%d %" FORMAT_int64 , &result->lives, &result->money ) != 2 )
        return NULL;
    
    /// Read the waves
    
    int wave_count;
    
    if( fscanf( in , "%d" , &wave_count ) != 1 )
        return NULL;
    
    if( wave_count < 0 )
        return NULL;
    
    for( int i = 0 ; i < wave_count ; i++ ){
        /**
         * Each wave's description
         *  is a list of pairs
         *  (mob type id, number of mobs from that type)
         */
        int pairs_count;
        int64 money_on_completion;
        
        if( fscanf( in , "%d %" FORMAT_int64 , &pairs_count, &money_on_completion ) != 2 )
            return NULL;
        
        result->money_on_completion.push_back( money_on_completion );
        result->all_waves.push_back( std::vector< std::pair<int,int> >() );
        
        for( int j = 0 ; j < pairs_count ; j++ ){
            int mob_type, count;
            
            if( fscanf( in , "%d %d" , &mob_type, &count ) != 2 )
                return NULL;
            
            if( mob_type < 0 or mob_type >= mob_count )
                return NULL; /// invalid wave?
            if( count < 0 )
                return NULL;
            
            result->all_waves.back().push_back( std::make_pair( mob_type , count ) );
        }
    }
    
    fclose( in );
    
    return result;
}
