#include <stdarg.h>
#include <ios>
#include <iostream>
#include <fstream>
#include <sstream>
#include <errno.h>

#include "main.hpp"
#include "Game.hpp"
#include "TextGame.hpp"
#include "TextGameReader.hpp"
#include "ManyRunsLogicInfo.hpp"

#include "test.hpp"

#define bprintf(msg, ... ) printf("%3d::" msg, depth, ##__VA_ARGS__ )

#define MAX_DEPTH 3
#define SIM_WAVES 2
#define MAX_WAVE_RETRIES 25 // LEKO
#define MAX_GOOD_TOWERS 9

TextGame *game;
unsigned long long total_sims, wave_sims;

ManyRunsLogicInfo *logicInfo = new ManyRunsLogicInfo;

unsigned int depth = 0;
unsigned int current_wave = 0;
int best_score, perfect_score;

// for all waves
std::deque< std::deque<std::string> > all_instructions;
std::deque< std::string > best_instruction; // for this wave

const std::vector< FieldCellCoord > no_upgrades; // empty vector

void game_runwave() {
	std::vector< RunningMob* > mobs;
	for (int i=0; i<(int)all_mobs[current_wave].size(); i++) mobs.push_back( new RunningMob(all_mobs[current_wave][i]->type));
	game->go( waveRunLogicManyInRangeCalls ,
			mobs ,
			defaultMobInRangeForManyCalls ,
			defaultMobFellDead ,
			defaultMobReachedTheEnd ,
			logicInfo
			);
}

/*----------------------------------------------*/

std::vector< RunningMob* > copymobs(int wave) {
	std::vector< RunningMob* > copy;
	for (int i=0; i<(int)all_mobs[wave].size(); i++) copy.push_back( new RunningMob(all_mobs[wave][i]->type));
	return copy;
}

// returns max level reached.lives left
int game_testwave(std::vector< RunningMob* > mobs) {
	int wave = current_wave;
	int lives = game->lives;
	int64 money = game->money;
	int score;

	if (wave > (int) game->all_waves.size()) return 0.0;


	//for (int i=0; i<(int)all_mobs[wave].size(); i++) mobs.push_back( new RunningMob(all_mobs[wave][i]->type));
	game->go( waveRunLogicManyInRangeCalls ,
			mobs ,
			defaultMobInRangeForManyCalls ,
			defaultMobFellDead ,
			defaultMobReachedTheEnd ,
			logicInfo
			);

	if( game->lives <= 0 ){
		//pprintf("Died at wave #%d", wave + 1);
	}

	score = game->lives;

	//printf("Survived wave %d (lives %d), got %" FORMAT_int64 " has %" FORMAT_int64 "\n", wave + 1, game->lives, game->money_on_completion[wave], game->money);
	//bprintf("sim done. wave=%d score=%.2f\n", wave, score);

	//restore
	game->lives = lives;
	game->money = money;

	return score;
}

int game_testwave_auto() {
	std::vector< RunningMob* > copy;
	copy = copymobs(current_wave);
	return game_testwave(copy);
}


// unsigned
uint64_t powi(int number, int exponent) {
	int i, product = 1;
	for (i = 0; i < exponent; i++) product *= number;
	return product;
}


std::string string_format(const std::string &fmt, ...) {
	int size = 10;
	std::string str;
	va_list ap;
	while (1) {
		str.resize(size);
		va_start(ap, fmt);
		int n = vsnprintf((char *)str.c_str(), size, fmt.c_str(), ap);
		va_end(ap);
		if (n > -1 && n < size) {
			str.resize(n);
			return str;
		}
		if (n > -1)
			size = n + 1;
		else
			size *= 2;
	}
	return str;
}

bool cb_good_sort_compare(const std::pair<int, int> a, const std::pair<int, int> b ) { return a.second > b.second; }

/// generate string instructions:

//sell
std::string gi_sell(TextGame *game, int x, int y) {
	std::string ret = string_format("5 %d %d\n", x, y);
	return ret;
}

// build
std::string gi_build(TextGame *game, int type, int *x, int *y, std::vector< FieldCellCoord > upgrades, upgrade_mode_e upgrade_mode, bool temp_tower) {
	std::string ret, cmd_buy, upgrade_coords, prereq;
	coverage *cov;
	int range, num_upgrade_towers;
	int phx=-1,phy=-1, upx=-1, upy=-1; // stack placeholder

	// check unused and fix range
	range = game->all_turrets[type]->range;

	if (NULL==x || NULL==y) {
		while (0 == (int) cached_coverage[range].size() && range > 0) {
			range--;
		}

		range = game->all_turrets[type]->range;
		while (0 == (int) cached_coverage[range].size() && range < 60) {
			range++;
		}

		if (0 == (int) cached_coverage[range].size()) {
			//not possible!?
			printf("no cached_coverage[%d]\n", range);
			exit(-111);
		}
		//printf("checking coverage for turret type %d = %d, result=%d/%d\n", type, game->all_turrets[type]->range, cov->x, cov->y);
		x = &phx; y = &phy; // stack placeholder
	}
	if (-1==*x || -1==*y) { // auto coords
		cov = cached_coverage[range][0];
		*x = cov->x;
		*y = cov->y;
		//printf("gi_build, auto coverage for %d = %d/%d\n", range, *x, *y);
	}
	ts_cached_coverage_remove_spot(*x, *y);
	// FIXME: check for free spot, 2x2 map has less than 64 build spots

	num_upgrade_towers = game->upgrade_graph[type].size();

	for (int i=0; i<num_upgrade_towers; i++) {
		//printf("type[%d] prereq[%d] = %d\n", type, i, game->upgrade_graph[type][i]);
	}

	cmd_buy = string_format("4 %d %d %d", *x, *y, type);

	// upgrades..
	switch (upgrade_mode) {
	case UP_AUTO:
		/* auto mode, big range for upgrades */
		// FIXME try with 2x1 maps and ..

#if 0
		for (int i=0; i<num_upgrade_towers; i++) {
			for (range=63; range>0; range--) {
				if (0 < cached_coverage[range].size()) break;
			}
			if (0 == range) {
				printf("no spot for upgrade tower\n");
				errno = EINVAL;
				return "";
			}
			cov = cached_coverage[range][0];
			upgrade_coords += string_format(" %d %d", cov->x, cov->y);
			ts_cached_coverage_remove_spot(cov->x, cov->y);
#endif
		for (int i=0; i<num_upgrade_towers; i++) {
			upx=-1; upy=-1;
			//printf("recursive[%d] gi_build ..", i);
			prereq += gi_build(game, game->upgrade_graph[type][i], &upx, &upy, no_upgrades, UP_AUTO, false);
			//printf("got tower @ %d/%d\n", upx, upy);
			upgrade_coords += string_format(" %d %d", upx, upy);
		}



		break;

	case UP_MANUAL:
		for (int i=0; i<(int)upgrades.size(); i++) {
			ret += string_format(" %d %d", upgrades[i].x, upgrades[i].y);
		}
		break;
	}

	ret = prereq + cmd_buy + upgrade_coords + "\n";
	//ret += "\n";

	return ret;
}

int gi_execute(std::deque< std::string > instructions, int mode) {
	int ins;
	int x,y, turret_id;
	int needed, r;
	std::stringstream stream (std::stringstream::in | std::stringstream::out);
	std::vector< FieldCellCoord > upgrades;

	//printf("execute[%d], go %d instructions\n", mode, instructions.size());
	for (int instr = 0; instr < (int) instructions.size(); instr++) {
		stream.str("");
		stream << instructions[instr];

		//printf("ins[%s]\n", stream.str().c_str());

		// scan instruction
		stream >> ins;
		if (stream.rdstate() != std::stringstream::goodbit) { printf("Err parsing instr code!\n"); return -1; }

		switch (ins) {
			case INSTRUCTION_ID_BUY:
				upgrades.erase(upgrades.begin(), upgrades.end());

				stream >> x >> y >> turret_id;
				if (stream.rdstate() != std::stringstream::goodbit) { printf("Err parsing instr data!\n"); return -1; }

				{ // get upgrades
					needed = game->upgrade_graph[turret_id].size();

					for( int j = 0 ; j < needed ; j++ ){
						int upgrade_x, upgrade_y;
						stream >> upgrade_x >> upgrade_y;
						if (stream.rdstate() != std::stringstream::goodbit) { printf("Err parsing upgrade data!\n"); return -1; }
#if 0 // gi_build() auto mode will use empty coords
						if( !game->field->isInside( upgrade_x , upgrade_y )
								or !game->field->isTurret( upgrade_x , upgrade_y ) ){
							printf("!!!!!Turret upgrade coords out-of-range / not turret's coords");
							return -1; // FIXME !?
						}
#endif
						upgrades.push_back( FieldCellCoord( upgrade_x , upgrade_y ) );
					}
				} // upgrades

				//printf("got instruction %d x/y=%d/%d tt=%d\n", ins, x, y, turret_id);

				switch (mode) {
				case IS_MODE_RUN:
					if (0 != needed) {
						/* doing upgrade */
						//1, recursive build dependancies
#if 0
						for (int i=0; i<needed; i++) {
							int id = game->upgrade_graph[turret_id][i];
							r = game->buyTurret(id, upgrades[i]->x, upgrades[i]->y, empty);
						}
#endif
					}

					//printf("buy turret %d cost%d total(%lld)  money %lld\n", turret_id, game->all_turrets[turret_id]->buy_price, total_tower_cost[turret_id], game->money);
					if((r=game->buyTurret( turret_id , x , y , upgrades)) < 0 ){
						printf("failed to buy turret: %d ins:%d free=%d, price:%d money:%lld total_price=%d)\n", r, game->field->isInside( x , y ), game->field->isFree( x , y ),
							 game->all_turrets[turret_id]->buy_price, game->money, total_tower_cost[turret_id]	);
						printf("tid = %d\n", game->field->getTurretId(x, y));
					}else{
						//printf("buy turret %d at (%d,%d) for %d turrets=%d\n", (int) turret_id, (int) x, (int) y, (int) game->all_turrets[turret_id]->buy_price, game->totalTurretsBuild());
						/// Set max waste to each new tower to 1.0
						logicInfo->maxWasteDamage[x][y] = 1.0;
						ts_map_turrets[x][y] = turret_id;
						for (int i=0; i<16; i++) {
							if (1 == double_compare(game->all_turrets[turret_id]->affect[i], 0)) mob_type_reorder[i] = 1;
						}
					}
					break;

				case IS_MODE_REVERT:
					game->money += game->all_turrets[turret_id]->buy_price;
					if (false == game->sellTurret(x, y, true)) {
						//printf("failed to sell turret: %d ins:%d free=%d, money:%lld total_price=%d(was it an upgrade ?)\n", r, game->field->isInside( x , y ), game->field->isFree( x , y ),
						//	 game->money, total_tower_cost[turret_id]	);
					}
					ts_map_turrets[x][y] = -1;

					break;
				} // switch (mode)

				break;

			case INSTRUCTION_ID_SELL:
				if (mode == IS_MODE_REVERT) {
					printf("INSTRUCTION_ID_SELL with IS_MODE_REVERT not implemented\n"); return -1;
				}
				stream >> x >> y;
				printf("sell turret @ %d/%d\n", x, y);
				if (stream.rdstate() != std::stringstream::goodbit) { printf("Err parsing sell data!\n"); return -1; }
				if (false == game->sellTurret(x, y, false)) {
					printf("failed to sell turret: %d ins:%d free=%d, money:%lld \n", r, game->field->isInside( x , y ), game->field->isFree( x , y ), game->money);
				}
				ts_map_turrets[x][y] = -1;

				break;


				break;
			default:
				printf("unhandled instruction %d\n", ins);
				exit(0);
		}

	}

}

#define TOFILE(out, dq) \
	for(int i=0; i<(int)dq.size(); i++) { \
		std::string s = dq[i]; \
		out << s; \
	}



void smarterTest(int argc, char **argv){
	//const char INPUT_FILE_NAME[] = "input.in";
	//const char OUTPUT_FILE_NAME[] = "output.out";
	const char *INPUT_FILE_NAME = argv[1];
	const char *OUTPUT_FILE_NAME = argv[2];
	int last_wave = 0;
	int total_upgrades = 0;
	std::deque<std::string> temp_wave_instr;

	static int got_helper[MAX_TURRET_COUNT];

	game = TextGameReader::read( INPUT_FILE_NAME );

	eprintf("Input %sOK", (game ? "" : "not "));
	if (NULL == game) {
		fprintf(stdout, "input error\n"); exit(-1);
	}

	std::ofstream outfile (OUTPUT_FILE_NAME, std::ios::out);


	// sod
	//ts_print_map(game);

	ts_init_coverage_test(game);
	ts_cache_cell_coverage(game);
	ts_cache_single_tower_score(game);
	ts_cache_mob_hp(game);

#if 1 // print
	for (int turret_id=0; turret_id<(int)game->all_turrets.size(); turret_id++) {
		fprintf(stderr, "SS%2d [", turret_id);
		for (int j=0; j<16; j++) fprintf(stderr, " %4.0f", ts_tower_score_single[turret_id][j]);
		fprintf(stderr, "]\n");
	}
#endif

#if 0
	ts_calc_tower_type_score(game, 0);
	printf("=========================\n");
	for (int i=0; i < (int) game->all_turrets.size(); i++) {
		printf("%d [%.0f]\n", i, ts_tower_score_total[i]);
	}
	printf("=========================\n");
#endif

	std::vector< FieldCellCoord > upgrades;

	/*----------------------------------------------*/

	// 1st pass, read all mobs
	for( int w = 0 ; w < (int) game->all_waves.size() ; w++ ){
		int wave_size = game->all_waves[w].size();

		double total_wave_hp = 0.0;
		for( int i = 0 ; i < wave_size; i++ ){
			int type = game->all_waves[w][i].first;
			int count = game->all_waves[w][i].second;
			double hp = (game->all_mobs[type]->health * (count * MOB_COUNT_MAGIC)) ;
			total_wave_hp += hp;
			double score = (hp * MOB_HP_MAGIC) + game->all_mobs[type]->speed * MOB_SPEED_MAGIC;
			score *= MOB_SCORE_MAGIC;
			//printf("mob%d x %d.spd%d. hp=%.1f score=%.1f\n", type, count, game->all_mobs[type]->speed, hp, score);
			wave_type_score[w][type] = score;

			for( int j = 0 ; j < count ; j++ )
				all_mobs[w].push_back( new RunningMob( game->all_mobs[type] ) );
		}
		wave_mob_types[w] = wave_size; //different mob types for this wave
		//printf("wave %d, turrets %d total wave hp: %.0f\n", w, game->_turrets_built, total_wave_hp);

#if 0		// print wave score
		printf("wave_type_score[");
		for (int i=0; i<16; i++) {
			if (wave_type_score[w][i] != 0) fprintf(stderr," %d:%.0f", i, wave_type_score[w][i]);
		}
		printf("]\n");
#endif
	} // for each wave


	/*----------------------------------------------*/


	total_sims = 0;
	int test_stop = 0;
	while (1) {
		std::vector< RunningMob* > mobs_copy;
		int update_coverage_cache = 1;
		int score;
		int survived_types[16];
		int wave_types[16];
		int good_tower[64], wave_builds_tower[64];
		std::deque<int> good_towers, bad_towers;

		perfect_score = game->lives;
		wave_sims = 0;


		printf("===========================current wave %d\n", current_wave);
		int towers_to_sell = game->totalTurretsBuild() / (MAX_WAVE_RETRIES/3) ;


		// retry point
		int wave_retries = MAX_WAVE_RETRIES;
retry_wave:
		best_score = -99999;
		update_coverage_cache = 1; // 1st time for this wave/retry
		best_instruction.erase(best_instruction.begin(), best_instruction.end());

		// test wave
		memset(survived_types, 0, sizeof(survived_types));
		memset(wave_types, 0, sizeof(wave_types));
		mobs_copy.erase(mobs_copy.begin(), mobs_copy.end());
		mobs_copy = copymobs(current_wave);
		score = game_testwave(mobs_copy);
		printf("pre .. wave %d score=%d\n", current_wave, score);

		for (int i=0; i<(int)mobs_copy.size(); i++) {
			if (mobs_copy[i]->type->health > EPS) survived_types[mobs_copy[i]->type->type]=1;
			wave_types[mobs_copy[i]->type->type] = 1;
		}

		printf("wave mob types[");
		for (int i=0; i<16; i++) {
			if (0 != wave_types[i]) printf(" %d", i);
		}
		printf("]\n");
		printf("survived types[");
		for (int i=0; i<16; i++) {
			if (0 != survived_types[i]) printf(" %d", i);
		}
		printf("]\n");


		// define good/bad towers
		good_towers.erase(good_towers.begin(), good_towers.end());
		bad_towers.erase(bad_towers.begin(), bad_towers.end());
		memset(good_tower, 0, sizeof(good_tower));
		for (int i=0; i<(int)game->all_turrets.size(); i++) {
			for (int type=0; type<16; type++) {
				if ( 0 != survived_types[type]
						&& game->all_turrets[i]->affect[type] > EPS
						&& 0 == good_tower[i] /* dup good tower */
						) {
					good_towers.push_back(i);
					good_tower[i] = 1;
				} else {
					//bad_towers.push_back(i);
				}
			}
		}

		// print
		printf("good["); for (int i=0; i<(int)good_towers.size(); i++) printf(" %d", good_towers[i]); printf("]\n");
		//printf(" bad["); for (int i=0; i<(int)bad_towers.size(); i++) printf(" %d", bad_towers[i]); printf("]\n");

#if 1
		// filter towers
		{
			int tower_frags[64]; // #kills for each type
			int old_lives = game->lives;
			int dead_mobs;
			memset(tower_frags, 0, sizeof(tower_frags));

			for (int i=0; i<(int)good_towers.size(); i++) {
				std::deque<std::string> wave_instructions;
				if (total_tower_cost[good_towers[i]] > game->money || tower_cost_towers[good_towers[i]] + game->totalTurretsBuild() >= (MAX_TURRET_COUNT-1)) continue;
				ts_cache_cell_coverage(game);

				std::string this_ins = gi_build(game, good_towers[i], NULL, NULL, no_upgrades, UP_AUTO, false);
#if 0
#if 1 // 38 -> 66s !?
				wave_instructions = buildstring_to_deque(this_ins);
#else
				std::deque<std::string> instr = buildstring_to_deque(this_ins);
				wave_instructions.insert(wave_instructions.end(), instr.begin(), instr.end());
#endif

#else
				if( this_ins.size() ) for( std::string::const_iterator i=this_ins.begin(), last=this_ins.begin(); i!=this_ins.end(); ++i ) {
					if( *i == '\n' ) { const std::string str(last,i+1); wave_instructions.push_back(str); last = i; ++last; }
				}
#endif

				gi_execute(wave_instructions, IS_MODE_RUN);

				dead_mobs = 0;
				mobs_copy.erase(mobs_copy.begin(), mobs_copy.end());
				mobs_copy = copymobs(current_wave);
				game_testwave(mobs_copy);

				gi_execute(wave_instructions, IS_MODE_REVERT);

				for (int j=0; j<mobs_copy.size(); j++) {
					if( 1 > double_compare(mobs_copy[j]->health, 0.0) ) dead_mobs++;
				}

				tower_frags[good_towers[i]] = dead_mobs;
				//printf("frags[%d] = %d\n", good_towers[i], dead_mobs);
			}


			{ // get topX
				std::deque< std::pair<int, int> > sortme;
				for (int i=0; i<(int)good_towers.size(); i++) sortme.push_back( std::make_pair(good_towers[i], tower_frags[good_towers[i]]) );
				std::sort(sortme.begin(), sortme.end(), &cb_good_sort_compare);

				good_towers.erase(good_towers.begin(), good_towers.end());
				for (int i=0; i<MAX_GOOD_TOWERS && i < sortme.size() && sortme[i].second > 0 ; i++) {
					//printf("use good[ tower #%d frags %d\n", sortme[i].first, sortme[i].second);
					if (sortme[i].first < game->all_turrets.size()) good_towers.push_back( sortme[i].first );
				}

				printf("good["); for (int i=0; i<(int)good_towers.size(); i++) printf(" %d(%d)", good_towers[i], sortme[i].second); printf("]\n");
			}

		}
#endif

/*----------------------------------------------*/

		unsigned long long perm_step = 0;
		unsigned long long bits = (int)good_towers.size();
		unsigned long long perm_max = powi(2, bits)-1;
		printf("total perms: %d\n", perm_max);

		while (perm_step <= perm_max) { // wave
			uint64_t  total_cost=0;
			std::deque<int> build_towers;
			std::deque<std::string> wave_instructions;

			//wave_instructions.erase(wave_instructions.begin(), wave_instructions.end());
			if (1 == update_coverage_cache) ts_cache_cell_coverage(game);
			update_coverage_cache = 0;

			int towers_to_build=0;
			for (int i=bits-1; i>=0; i--) {
				if (0 != (perm_step & (1<<i))) {
					towers_to_build += tower_cost_towers[good_towers[i]];
					build_towers.push_back(good_towers[i]);
					total_cost += total_tower_cost[good_towers[i]];
				}
			}

			if (towers_to_build + game->totalTurretsBuild() >= (MAX_TURRET_COUNT-1)) { // allow max 63 towners (need 1 for pre wave test)
				//printf("ttb:%d + total %d = %d\n", towers_to_build, game->totalTurretsBuild(), towers_to_build + game->totalTurretsBuild());
				// skip this perm. cant build more
				goto next_permutation;
			}

			if (total_cost > game->money) {
				// skip this perm. not enough money
				goto next_permutation;
			}


			for (int i=0; i<(int)build_towers.size(); i++) { // build string into wave_instructions
				std::string this_ins = gi_build(game, build_towers[i], NULL, NULL, no_upgrades, UP_AUTO, false);
				std::deque<std::string> instr = buildstring_to_deque(this_ins);
				wave_instructions.insert(wave_instructions.end(), instr.begin(), instr.end());
			}

			update_coverage_cache = 1;
			memset(mob_type_reorder, 0, sizeof(mob_type_reorder));
			gi_execute(wave_instructions, IS_MODE_RUN);

			for (int i=0; i<16; i++) { // for each type
				if (1 == mob_type_reorder[i]) ts_set_type_order(game, i, logicInfo);
			}

			// run test
			mobs_copy.erase(mobs_copy.begin(), mobs_copy.end());
			mobs_copy = copymobs(current_wave);

			score = game_testwave(mobs_copy);

			if (0 == (perm_step % 256)) {
				printf("[%4d] do perm: %4lld/%4lld, tc=%5lld score=%3d ", current_wave, perm_step, perm_max, total_cost, score);
				printf(" / %3d / %3d\n", best_score, perfect_score);
			}


			// if score == perfect_score just run wave
			if (score > best_score) { // new best score, record and continue
				printf("new best score %d -> %d. %d instructions\n", best_score, score, wave_instructions.size());
				best_score = score;
#if 0
				for (int i=0; i<(int)wave_instructions.size(); i++) {
					best_instruction.push_back(wave_instructions[i]);
				}
#else
				best_instruction = wave_instructions;
#endif
			}

			// undo
			//printf("before revert. turrets=%d, is size=%d\n", game->totalTurretsBuild(), wave_instructions.size());
			gi_execute(wave_instructions, IS_MODE_REVERT);
			//printf("after revert. turrets=%d, is size=%d\n", game->totalTurretsBuild(), wave_instructions.size());


			if (score == perfect_score) {
				printf("perfect score with perm %lld\n", perm_step);
				break;
			}

			//printf("\n");
next_permutation:
			perm_step++;
		} // while this wave
		//printf("perm loop done. %lld/%lld\n", perm_step, perm_max);

		memset(mob_type_reorder, 0, sizeof(mob_type_reorder));
		// execute best instr
		printf("executing best (so far)..size=%d\n", best_instruction.size());
		gi_execute(best_instruction, IS_MODE_RUN);

		temp_wave_instr.erase(temp_wave_instr.begin(), temp_wave_instr.end());
		temp_wave_instr.insert(temp_wave_instr.end(), best_instruction.begin(), best_instruction.end());
#if 1
		for (int i=0; i<16; i++) { // for each type
			if (1 == mob_type_reorder[i]) ts_set_type_order(game, i, logicInfo);
		}
#endif

		ts_print_map(game);

		// za da mozhe da postori poveche ot 1 kula ot syshtia tip
		if (score != perfect_score && wave_retries > 0) {
			if (0 != best_instruction.size()) {
				printf("finished wave with imperfect score, retry %d\n", wave_retries);
				wave_retries--;
				goto retry_wave;
			} else {
				printf("nothing to do, sell towers %d\n", wave_retries);
				{ /** sell here */
					std::deque< std::string > instructions; // for this wave
					int sold=0, runs=20;
					memset(wave_builds_tower, 0, sizeof(wave_builds_tower));

					while (sold < towers_to_sell && runs-->0) {
						for (int i=0; i<(int)ts_sorted_tower_score.size() && sold < towers_to_sell; i++) {
							if (1 == good_tower[ts_sorted_tower_score[i].first]) continue; // do not sell 'good' towers
							for (int j=0; j<64; j++) {
								if (NULL == game->turrets[j]) continue;
								const PlacedTurret *pt = game->turrets[j];
								if (-1 == ts_map_turrets[pt->place.x][pt->place.y]) continue;
								if (ts_sorted_tower_score[i].first == ts_map_turrets[pt->place.x][pt->place.y] ) {
									std::string this_ins = gi_sell(game, pt->place.x, pt->place.y);
									instructions.push_back(this_ins);
									sold++;
									ts_map_turrets[pt->place.x][pt->place.y] = -1;
									break;
								}
							} //for j
						} //for i
					} // while

					int64 oldmoney = game->money;
					// execute
					gi_execute(instructions, IS_MODE_RUN);
					temp_wave_instr.insert(temp_wave_instr.end(), instructions.begin(), instructions.end());
					printf("sold %d towers, earned %lld\n", instructions.size(), game->money - oldmoney);
				}
				wave_retries--;
				goto retry_wave;
			}
		}

		// run wave
		game_runwave();
		if (game->lives <= 0) {
			printf("Died! wave=%d turrets=%d money=%lld\n", current_wave, game->totalTurretsBuild(), game->money);
			ts_print_turret_report(game);
			break;
		}

next_wave:
		//FIX output instructions
		game->money += game->money_on_completion[current_wave];
		printf("wave %d done, lives=%d money=%lld turrets=%d\n", current_wave, game->lives, game->money, game->totalTurretsBuild());

		// file
		outfile << temp_wave_instr.size() << std::endl;
		TOFILE(outfile, temp_wave_instr);

		current_wave++;

		if ((int) game->all_waves.size() == current_wave) {
			printf("survived all waves!\n");
			break;
		}
		if (++test_stop == 9999) break;
	}


	// TODO 0 instructions for other waves
	for (int i=current_wave; i<(int)game->all_waves.size(); i++) {
		outfile << 0 << std::endl;
	}

	outfile.close();


	return;
}
/*----------------------------------------------*/

void go(int argc, char **argv){
	//stupidTest();

	smarterTest(argc, argv);
}

int main(int argc, char **argv){
	if (3 != argc) {
		printf("usage: %s <in file> <out file>\n", argv[0]);
		exit(-1);
	}
	srand(time(0));// ^ getpid());
	//close(1); close(2);
	go(argc, argv);

	return 0;
}



