#ifndef MAIN_HPP_INCLUDED
#define MAIN_HPP_INCLUDED

#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <cstring>
#include <queue>
#include <ctime>
#include <cassert>
#include <inttypes.h>
#include <cstdio>
#include <cmath>
#include <utility>
#include <algorithm>

#define foreach(_var,_container) for( typeof( (_container).begin() ) _var = (_container).begin() ; _var != (_container).end() ; ++_var )
#define now() double( double( clock() ) / double( CLOCKS_PER_SEC ) )
//#define now() 0.0

/// Warning: if on unix os change '\\' to '/' in the call to strrchr()
/// To turn debugging off set DEBUG_LEVEL to 0

//#define OS_PATH_SEPARATOR '\\'
#define OS_PATH_SEPARATOR '/'

/// Change to corresponding for your OS
/// That's for WINDOWS, for Unix (I think it is) llu
//#define FORMAT_int64 "I64u"
#define FORMAT_int64 "llu"

#define MY_ERROR_DIE(points , msg) { printf("%d\n%s", points, msg); return; }
#define MY_OK_DIE(points , msg) { printf("OK %d\n%s", points, msg); return; }

#define MY_CHECK_FEOF(file) { if( feof(file) ){ MY_ERROR_DIE( 0 , "Invalid format" ); } }

#define DEBUG_LEVEL 2

#if DEBUG_LEVEL >= 2
#define eprintf(msg, ... ) fprintf(stderr," %s:%d in %s at %.4f :: " msg "\xA" , __FILE__ , __LINE__ , __FUNCTION__ , (float) now() , ##__VA_ARGS__ )
#else
#define eprintf(msg, ... ) 0
#endif

#if DEBUG_LEVEL >= 1
//#define pprintf(msg, ... ) fprintf(stderr," %s:%d in %s at %.4lf :: " msg "\xA" , strrchr( __FILE__ , OS_PATH_SEPARATOR )+1 , __LINE__ , __FUNCTION__ , now() , ##__VA_ARGS__ )
#define pprintf( msg , ... ) fprintf( stderr , " %s:%d in %s at %.4f :: " msg "\xA" , __FILE__, __LINE__ , __FUNCTION__ , (float) now() , ##__VA_ARGS__ )
#else
#define pprintf(msg, ... ) 0
#endif

const int MAX_TURRET_COUNT = 64;
//const int MAX_MOB_COUNT = 512;
const int MAX_MOB_TYPES = 16;
const int MAX_FIELD_SIZE = 64;
const int MAX_TURRET_NAME_LENGTH = 32;
const int MAX_INSTRUCTIONS_PER_WAVE = 1024;

const int INSTRUCTIONS_REQUIRES_COORDS = 2;
const int INSTRUCTION_ID_SET_ORDER = 1;
const int INSTRUCTION_ID_SET_MAX_WASTE = 3;
const int INSTRUCTION_ID_BUY = 4;
const int INSTRUCTION_ID_SELL = 5;

const int LAST_VALID_INSTRUCTION_ID = 5;

const int DELTA_X[] = { 0 , 0 , -1 , +1 };
const int DELTA_Y[] = { -1 , +1 , 0 , 0 };

const double EPS = 1e-9;

const int DIRECTIONS = 4;

typedef unsigned long long int64;

const int64 POW2[] =
{
1ULL, // 2^0
2ULL, // 2^1
4ULL, // 2^2
8ULL, // 2^3
16ULL, // 2^4
32ULL, // 2^5
64ULL, // 2^6
128ULL, // 2^7
256ULL, // 2^8
512ULL, // 2^9
1024ULL, // 2^10
2048ULL, // 2^11
4096ULL, // 2^12
8192ULL, // 2^13
16384ULL, // 2^14
32768ULL, // 2^15
65536ULL, // 2^16
131072ULL, // 2^17
262144ULL, // 2^18
524288ULL, // 2^19
1048576ULL, // 2^20
2097152ULL, // 2^21
4194304ULL, // 2^22
8388608ULL, // 2^23
16777216ULL, // 2^24
33554432ULL, // 2^25
67108864ULL, // 2^26
134217728ULL, // 2^27
268435456ULL, // 2^28
536870912ULL, // 2^29
1073741824ULL, // 2^30
2147483648ULL, // 2^31
4294967296ULL, // 2^32
8589934592ULL, // 2^33
17179869184ULL, // 2^34
34359738368ULL, // 2^35
68719476736ULL, // 2^36
137438953472ULL, // 2^37
274877906944ULL, // 2^38
549755813888ULL, // 2^39
1099511627776ULL, // 2^40
2199023255552ULL, // 2^41
4398046511104ULL, // 2^42
8796093022208ULL, // 2^43
17592186044416ULL, // 2^44
35184372088832ULL, // 2^45
70368744177664ULL, // 2^46
140737488355328ULL, // 2^47
281474976710656ULL, // 2^48
562949953421312ULL, // 2^49
1125899906842624ULL, // 2^50
2251799813685248ULL, // 2^51
4503599627370496ULL, // 2^52
9007199254740992ULL, // 2^53
18014398509481984ULL, // 2^54
36028797018963968ULL, // 2^55
72057594037927936ULL, // 2^56
144115188075855872ULL, // 2^57
288230376151711744ULL, // 2^58
576460752303423488ULL, // 2^59
1152921504606846976ULL, // 2^60
2305843009213693952ULL, // 2^61
4611686018427387904ULL, // 2^62
9223372036854775808ULL, // 2^63
};

inline int double_compare( const double a , const double b ){
    if( fabs(a - b) < EPS )
        return 0;
    else if( a + EPS < b )
        return -1;
    else
        return +1;
}

struct WaveInfo;
struct RunningMob;
struct PlacedTurret;
struct Game;
struct WaveRunner;

typedef void (*WaveCompleted_Callback_Type)( WaveInfo *result , Game *game );
typedef void (*WaveRunner_Callback_Type)( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument );
typedef void (*WaveRunner_Logic_Callback)( WaveRunner *waveRunner , const std::vector< RunningMob* > &mobs , Game &game , WaveInfo *result , void *argument );

#endif // MAIN_HPP_INCLUDED
