#include "test.hpp"
#include "TextGame.hpp"
#include "Game.hpp"

char turrets_array[] = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ[]{}()";
unsigned char ts_map_turrets[64][64]; // store turret ids (to draw map)
unsigned int mob_hp[16]; // for every mob type

// wave_type score
double wave_type_score[3000][16];
unsigned int wave_mob_types[3000]; // mob types for all waves
uint64_t total_tower_cost[64]; // total. tower price + ugprades
uint32_t tower_cost_towers[64]; //tower price in towers
int mob_type_reorder[16];

// mobs for all waves
std::vector< RunningMob* > all_mobs[3000];

void ts_print_map(TextGame *game ) {
	int i, j;

	printf("== map\n");
	for (i=0; i<game->field->rows; i++) {
		for (j=0; j<game->field->cols; j++) {
			switch(game->field->cell[i][j].type) {
				case FIELD_PATH:
					printf("*");
					break;
				case FIELD_BLOCKED:
					printf("!");
					break;
				case FIELD_FREE:
					printf(" ");
					break;
				case FIELD_TURRET:
					//int turret_seq = turrets_array[game->field->cell[i][j].turret];
					int turret_seq = turrets_array[ ts_map_turrets[i][j] ];
					printf("%c", turret_seq);
					break;

			}

		}
		printf("\n");
	}
	printf("== map end\n");

}

void ts_print_turrets_info(const Game *game) {
	int i, j;

	printf("== TI\n");
	for (i=0; i<64; /*game->all_turrets.size();*/ i++) {
 		const PlacedTurret *t = game->getTurret( i );
		if (NULL == t) break;
		printf("id: %d (%d,%d) t(%d) f/k[%d/%d]",
				i, t->place.x, t->place.y, t->type->type,
				t->fires, t->kills);
		printf(" name[%s] rate[%d], range[%d]\n",
				t->type->name.c_str(), t->type->rate, t->type->range);

		printf("damage:");
		for (j=0; j<MAX_MOB_TYPES; j++) {
			printf("%.2f ", t->type->affect[j]);
		}
		printf("\n");

	}
	printf("== TI end\n");

}

double ts_rate_formula(double rate) {
	if (rate < 0) return (-rate) * NEGATIVE_RATE_MAGIC;
	return rate;
}

/*----------------------------------------------*/
coverage::coverage(int x, int y, int cov) : x(x), y(y), cov(cov) {} ;
bool ts_cb_coverage_compare (const coverage *a, const coverage *b ) { return a->cov > b->cov; } 

std::deque<coverage *> cached_coverage[64];
// caches turrent path coverage
void ts_cache_cell_coverage(const Game *game) {
	int range, i, j, r;
	coverage *cov;

	for (range=1; range<64; range++) {
		cached_coverage[range].clear();
		for (i=0; i<game->field->rows; i++) {
			for (j=0; j<game->field->cols; j++) {
				if (FIELD_FREE != game->field->cell[i][j].type) continue;
#if 0
				if (game->field->isTurret(j, i)) {
					printf("isturret!\n"); exit(12);
				}
#endif
				r = ts_cell_test_coverage(game, j, i, range);
				if (0 == r) continue;
				cov = new coverage(i, j, r);
				cached_coverage[range].push_back(cov);
			}
		}
	}

	// sort all
	for (range=1; range<64; range++) {
		//std::sort(cached_coverage[range].begin(), cached_coverage[range].end());
		std::sort(cached_coverage[range].begin(), cached_coverage[range].end(), ts_cb_coverage_compare);
	}
}
bool coverage::operator<( const coverage &a ) const { printf("!!!"); return this->cov > a.cov; } 

// clear spot from all range caches
void ts_cached_coverage_remove_spot(int x, int y) {
	int range, i;
	for (range=1; range<64; range++) {
		for (i=0; i<(int)cached_coverage[range].size(); i++) {
			if (x==cached_coverage[range][i]->x && y==cached_coverage[range][i]->y) {
				//printf("remove from range[%d] (%d) x/y=%d/%d\n", range, i, x, y);
				coverage *cov = cached_coverage[range][0];
				cached_coverage[range].erase( cached_coverage[range].begin() + i);
				//delete cov; // FIXME !?
				continue; // done with this range
			}
		}
	}
}

int ts_cached_cell_coverage[64][64][64]; // [range] [x] [y]


unsigned long long total_tcost(TextGame *game, int turret_id, int *num_towers) {
	if (turret_id >= game->all_turrets.size()) return 0ull;
	if (NULL == game->all_turrets[turret_id]) return 0ull;
	unsigned long long ret  = game->all_turrets[turret_id]->buy_price;
	//printf("!%d=%lld ", turret_id, ret);
	for (int i=0; i<(int) game->upgrade_graph[turret_id].size(); i++) {
		ret += total_tcost(game, game->upgrade_graph[turret_id][i], num_towers);
		//printf("+%lld ", total_tcost(game, i));
	}
	//printf("\n");
	*num_towers += 1;
	return ret;
}

int ts_init_coverage_test(TextGame *game) {
	int i, j, ret, x, y;

	for (int range=0; range<64; range++) {
		for (y=0; y<game->field->rows; y++) {
			for (x=0; x<game->field->cols; x++) {
				ret = 0;

				i = x-range >= 0 ? x-range : 0;
				j = y-range >= 0 ? y-range : 0;
				//printf("rows/cols %d/%d\n", game->field->rows, game->field->cols);
				//printf("loop1 %d..%d, %d..%d\n", i, x+range, j, y+range);

				for (j = y-range >= 0 ? y-range : 0; j<=y+range &&  j<game->field->rows; j++) {
					for (i = x-range >= 0 ? x-range : 0; i<=x+range && i<game->field->cols; i++) {
						switch(game->field->cell[j][i].type) {
							case FIELD_PATH:
								//printf("	+(%d,%d)\n", i, j);
								ret++;
								break;
						}
					}
				}
				//printf("cached_cell_coverage[%d][%d][%d] = %d\n", range, x, y, ret);
				ts_cached_cell_coverage[range][x][y] = ret;
			}
		}
	}

	// also init total towers cost
	
	for (int i=0; i<64; i++) {
		int num_towers=0; // price in #towers
		total_tower_cost[i] = total_tcost(game, i, &num_towers);
		tower_cost_towers[i] = num_towers;
		if (0 != total_tower_cost[i]) printf("total tower cost[%d] = %lld(%d)\n", i, total_tower_cost[i], tower_cost_towers[i]);
	}
}

std::deque<std::string> buildstring_to_deque(std::string bs) {
	std::deque<std::string> ret;
	//printf("got[%s]\n", this_ins.c_str());
	if( bs.size() ) {
		std::string::const_iterator last = bs.begin();
		for( std::string::const_iterator i=bs.begin(); i!=bs.end(); ++i ) {
			if( *i == '\n' ) {
				const std::string str(last,i+1);
				//printf("split push [%s]\n", str.c_str());
				ret.push_back(str);
				last = i;
				++ last;
			}
		}
	}

	return ret;
}

// query path cells in range
//!!! vzima Y, X
int ts_cell_test_coverage(const Game *game, int x, int y, int range) {
	return ts_cached_cell_coverage[range][x][y];

}

bool cb_towerscore_compare(const std::pair<int, double> a, const std::pair<int, double> b ) { 
	if (-1 == double_compare(a.second, b.second)) return true;
	return false;
} 

// type score
// for multi mob waves, divide damage
double ts_tower_score[16]; // sum of damage
double ts_tower_score_total[16];
double ts_tower_score_wave[16];
double ts_tower_score_single[64][16]; // solo score
std::deque< std::pair<int, double> > ts_sorted_tower_score;

// fills tower solo score
void ts_cache_single_tower_score(TextGame *game) {
	double coverage_formula = 1.0;
	for (int turret_id=0; turret_id<(int)game->all_turrets.size(); turret_id++) {
		double dmg = 0.0;
		const Turret *t = game->all_turrets[turret_id];
		for (int j=0; j<16; j++) {
			ts_tower_score_single[turret_id][j] = (ts_rate_formula(t->rate) * coverage_formula * (t->affect[j]));
			dmg += t->affect[j];
		}
		ts_tower_score[turret_id] = dmg;
	}

	// sort
	for (int i=0; i<(int)game->all_turrets.size(); i++) ts_sorted_tower_score.push_back( std::make_pair(i, ts_tower_score[i]) );
	std::sort(ts_sorted_tower_score.begin(), ts_sorted_tower_score.end(), &cb_towerscore_compare);

#if 1 // print
	for (int i=0; i<(int)game->all_turrets.size(); i++) {
		printf("tower score [%2d] = %.1f\n", ts_sorted_tower_score[i].first, ts_sorted_tower_score[i].second);
		//printf("SS%2d [", turret_id);
		//for (int j=0; j<16; j++) printf(" %4.0f", ts_tower_score_single[turret_id][j]);
		//printf("]\n");
	}
#endif
}

void ts_cache_mob_hp(TextGame *game) {
	for (int j=0; j<16; j++) {
		Mob *m = game->all_mobs[j];
		mob_hp[j] = m->health;
	}
}

void ts_calc_tower_type_score(TextGame *game, int wave) {
	int i, j;
	const Turret *t;
	const PlacedTurret *pt;
	unsigned int coverage;
	//double tower_coef[64][16];

	for (j=0; j<16; j++) {
		// reinit
		ts_tower_score_total[j] = 0.0;
		ts_tower_score_wave[j] = 0.0;
	}

	// balacne for this wave
	//for (i=0; i<(int)game->all_turrets.size(); i++) {
	for (i=0; i<game->totalTurretsBuild(); i++) {
		double total_damage = 0;
		double wave_damage = 0;
		int affected;
		pt = game->turrets[i];
		t = pt->type;

		coverage = ts_cell_test_coverage(game, pt->place.y, pt->place.x, pt->type->range);
		// FIXME! coverage does not matter, game engine will fire at max rate if it sees mob
		//coverage = 1; // 2pass towers still count

		for (j=0; j<16; j++) {

			total_damage += t->affect[j];
			//if (t->affect[j] > EPS) { // deals damage for this type
			if (t->affect[j] > EPS && wave_type_score[wave][j] > 1/*FIXME?*/) { // deals damage for this type
				wave_damage += t->affect[j];
				affected++; // types affected
			}
		}

		// 2nd pass, fix damage rating
		for (j=0; j<16; j++) {
			double coverage_formula = (double) coverage / (double)game->all_mobs[j]->speed;
		//if (wave==1 && j==1 && t->affect[j] > 1) asm("int $3");
			double type_percent = (t->affect[j] / (wave_damage > EPS ? wave_damage : 1.0) ) * (wave_type_score[wave][j] > EPS ? 1.0 : 0.0);
			//double hits_to_kill = mob_hp[j] / (t->affect[j] > EPS ? t->affect[j] : mob_hp[j]);
			double hits_to_kill = mob_hp[j] < t->affect[j] ? 1 : mob_hp[j] / (t->affect[j] > EPS ? t->affect[j] : mob_hp[j]);
			double est_kills = t->rate / hits_to_kill;
			//double hits_to_kill_rebalance = (double) est_kills / t->rate;
			double hits_to_kill_rebalance = est_kills / t->rate;

#if 0 // test
			if (1 || wave==35 && i ==16) {
			 fprintf(stderr,  "type%2d hp%3d damage=%.1f rate=%d HTK=%.1f EK=%.1f rebalance = %.2f\n", j, mob_hp[j], t->affect[j], t->rate, hits_to_kill, est_kills, hits_to_kill_rebalance);
			}
#endif
			double score;
			type_percent *= TYPE_PERCENT_MAGIC;

#if 0 /* test unbalanced mob/tower? score */
			double this_damage = (ts_rate_formula(t->rate) * coverage_formula * (t->affect[j] * type_percent));
			if (wave==17 && j==3 ) 
				printf("T%d wave_score %.0f += %.0f\n", i, ts_tower_score_wave[j], this_damage); // total score
#endif
			score = (ts_rate_formula(t->rate) + coverage_formula + (t->affect[j])); // total score
			ts_tower_score_total[j] += score * hits_to_kill_rebalance;
#if 0
			double type_damage = t->rate * coverage_formula * (t->affect[j] * type_percent);
			if (type_damage > EPS) {
				printf("affect(%.0f) wave_damage(%.1f) rate(%d) cov(%d) type%%(%.0f)\n", t->affect[j], wave_damage, t->rate, coverage_formula, type_percent);
					printf("wave_score[%d] + %.f\n", j, type_damage);
			}
#endif
			score = (ts_rate_formula(t->rate) * coverage_formula * (t->affect[j] * type_percent)); // score for this wave
			ts_tower_score_wave[j] += score * hits_to_kill_rebalance;
			// FIXME or (rate * cov * affect) * type_percent ??

		}
	}

	return;
}

void ts_calc_total_tower_score(TextGame *game) {
	//for (int i=0; i < (int) game->all_turrets.size(); i++) {

}

// return [id] best turret i can buy for mob_type
int ts_get_bestturret_type(const TextGame *game, int mob_type, int64 max_price, int max_upgrades) {
	int affect_towerid = -1;
	float affect_max = 0;
	const Turret *t;
	coverage *cov;
	int64 buy_price;

	if (0 == max_price) max_price = game->money;

	// pick turret for mob_type
	for (int i=0; i<(int)game->all_turrets.size(); i++) {
		int upgrades_count = game->upgrade_graph[i].size();
		if (upgrades_count + 1 + game->totalTurretsBuild() > 64) continue; // no space for upgrades
		t = game->all_turrets[i];
		buy_price = t->buy_price;

		cov = cached_coverage[t->range][0];

		//printf("turret[%s] affect[%d]=%.2f\n", t->name.c_str(), mob_type, t->affect[mob_type]);
		if (cov->cov > 0 && affect_max < t->affect[mob_type] && upgrades_count <= max_upgrades) {
			for (int j=0; j<upgrades_count; j++) buy_price += game->all_turrets[ game->upgrade_graph[i][j] ]->buy_price;


			if (buy_price > game->money) { // check if i have money
				//printf("cant afford [%s] (%lld) > cash(%lld) UP(%d)\n", t->name.c_str(), buy_price, game->money, upgrades_count);
			} else {
				//printf("mob type %d max affect=%.2f\n", mob_type, t->affect[mob_type]);
				affect_max = t->affect[mob_type];
				affect_towerid = i;
			}
		}
	}
	return affect_towerid;
}

// buy turret, 1 = OK
FieldCellCoord ts_buy_turret(TextGame *game, ManyRunsLogicInfo *loficinfo, int turret_id, std::vector< FieldCellCoord > upgrades) {
	const Turret *t;
	int r;
	ManyRunsLogicInfo *logicInfo = new ManyRunsLogicInfo;

	t = game->all_turrets[turret_id];

	// best position
	coverage *best_slot = cached_coverage[t->range][0];
	int x=best_slot->x, y=best_slot->y;
	ts_cached_coverage_remove_spot(x, y); // remove from cache list
	r = game->buyTurret( turret_id , x , y , upgrades );
	if( r < 0 ){
		printf("Failed to buy turret(%d)\n", r);
		//asm("int $3");
	} else {
		logicInfo->maxWasteDamage[x][y] = 1.0;
	}
	ts_map_turrets[x][y] = turret_id; // store map
	return FieldCellCoord(x, y);
}

void bye(TextGame *game, int wave, int total_upgrades) {
	printf("%d, w: %4d tb: %2d l: %4d $:%10lld tUP:%4d\n", wave, wave, game->totalTurretsBuild(), game->lives, game->money, total_upgrades);
	exit(wave);
}

// set tower order for type
void ts_set_type_order(TextGame *game, int mob_type, ManyRunsLogicInfo *logicInfo) {
	static int got_helper[MAX_TURRET_COUNT];
	std::vector< int > order;
	std::deque <int> dq; // dq order
	PlacedTurret *pt;
	int turret_type;

	double del_damages[64]; // test..

	// populate order
	for (int turret_id=0; turret_id<64; turret_id++) {
		if (NULL == game->turrets[turret_id]) continue;
		pt = game->turrets[turret_id];
		turret_type = ts_map_turrets[pt->place.x][pt->place.y];
		//del_damages[turret_id] =  ts_tower_score_single[turret_type][mob_type];
		if (ts_tower_score_single[turret_type][mob_type] > EPS) { // deals damage
			//printf("front(%d)t%d[%.1f]\n", turret_id, mob_type, ts_tower_score_single[turret_type][mob_type]);
			dq.push_front(turret_id);
		} else {
			dq.push_back(turret_id);
		}
	}

#if 0
	printf("==order(%2d)[", mob_type);
	for (int i=0; i<(int)dq.size(); i++) printf(" %2d", dq[i]);
	//for (int i=0; i<(int)dq.size(); i++) printf(" %d(%.0f)", dq[i], del_damages[dq[i]]);
	printf("]==done(%2d)\n", mob_type);
#endif

	for (int i=0; i<(int)dq.size(); i++) order.push_back(dq[i]); // typo.. copy

	//order.push_back( game->field->getTurretId( turret_x , turret_y ) );

	/// Fill the permutation...
	//std::fill( got_helper , got_helper + MAX_TURRET_COUNT , 0 ); // memset?
	memset(&got_helper, 0, sizeof(got_helper));

	for( int j = 0 ; j < game->totalTurretsBuild() ; j++ ){
		if( got_helper[ order[j] ] ){
			eprintf("Same turret appears twice in an ordering j=%d o(j)=%d, help( o(j) )=%d\n", j, order[j], got_helper[ order[j] ] );
			exit(-23); // FIXME leko
			break;
		}
		//eprintf("OK ordering j=%d o(j)=%d, help( o(j) )=%d\n", j, order[j], got_helper[ order[j] ] );
		got_helper[ order[j] ]++;
	}

	// max 64 kuli, za vsiaka ne postroena push(id)
	for( int j = 0 ; j < MAX_TURRET_COUNT ; j++ )
		if( got_helper[j] == 0 )
			order.push_back( j );

	if( !logicInfo->setManualOrdering( mob_type , order ) ){
		eprintf("Failed to set manual order for %d (not a permutation?)", mob_type);

		MY_ERROR_DIE( 0 , "Setting manual ordering failed" );
	}else{
		//eprintf("Set manual order for %d", mob_type);
	}

}


void ts_print_turret_report(TextGame *game) {
	unsigned int turret_type[64];
	memset(turret_type, 0, sizeof(turret_type));
	for (int t=0; t<64; t++) {
		if (NULL == game->turrets[t]) continue;
		const PlacedTurret *pt = game->turrets[t];
		int type = ts_map_turrets[pt->place.x][pt->place.y];
		turret_type[type]++;
	}


	printf("=== tower report\n");
	for (int i=0; i<64; i++) {
		if (turret_type[i] ==0) continue;
		printf("type[%d] x %d\n", i, turret_type[i]);
	}

}
